render_types = [
    ('dropdown', 'Dropdown'),
    ('file', 'File'),
    ('hidden', 'Hidden'),
    ('number', 'Number'),
    ('radio', 'Radio'),
    ('text', 'Text'),
    ('text-file', 'Text File'),
    ('toggle', 'Toggle'),
]


def validate_argument(argument_data):
    error_dict = {}
    key = validate_key(argument_data, error_dict)
    if key is None:
        return error_dict

    error_dict[key] = {}
    argument_error_dict = error_dict[key]

    validate_unsupported_argument_fields(key, argument_data, argument_error_dict)

    # Recursively validate sub_arguments
    sub_arguments = argument_data.get('sub_arguments', {})
    for name, sub_arguments in sub_arguments.items():
        for subargument in sub_arguments:
            errors = validate_argument(subargument)
            if errors:
                if 'sub_arguments' not in error_dict:
                    argument_error_dict['sub_arguments'] = [errors]
                else:
                    argument_error_dict['sub_arguments'].append(validate_argument(subargument))

    validate_required(key, argument_data, argument_error_dict)
    type = validate_type(key, argument_data, argument_error_dict)

    if not type:
        return error_dict

    validate_description(key, argument_data, type, argument_error_dict)

    # Just return empty dict if we have no errors
    if error_dict[key]:
        return error_dict
    else:
        return {}


def validate_key(argument_data, error_dict):
    if 'key' not in argument_data.keys():
        error_dict['required'] = [
            f'One of your arguments is missing a key. Please specify a key for each of your arguments'
        ]
        return None
    else:
        return argument_data['key']


def validate_required(key, argument_data, error_dict):
    if 'required' in argument_data.keys():

        if not isinstance(argument_data['required'], bool):
            error_dict['required'] = [
                f'Invalid value in required specified on {key} argument. required can be true or false'
            ]


def validate_type(key, argument_data, error_dict):
    # Type is optional
    if 'type' in argument_data.keys():

        render_types_choices = [type_tuple[0] for type_tuple in render_types]
        type = argument_data['type']
        if type not in render_types_choices:
            error_dict['type'] = [
                f'Invalid value {type} in type specified on {key} argument \
                type can be one of {render_types_choices}'
            ]
            return ''

        if type == 'toggle':
            if 'options' not in argument_data:
                error_dict['type'] = [
                    f'There must be exactly 2 options ("on" and "off") on arguments of type toggle'
                ]
                return ''

            number_of_options = len(argument_data['options'].keys())

            if number_of_options != 2:
                error_dict['type'] = [
                    f'There must be exactly 2 options ("on" and "off") on arguments of type toggle. Received \
                    {number_of_options} options'
                ]
                return ''

            option_names = list(argument_data['options'].keys())

            if option_names not in (['on', 'off'], ['off', 'on']):
                error_dict['type'] = [
                    f'The two options on arguments of type toggle must be named "on" and "off". Received \
                    {", ".join(option_names)}'
                ]
                return ''

        return type


def validate_description(key, argument_data, type, error_dict):
    if 'description' not in argument_data.keys() and type != 'hidden':
        error_dict['argument_description'] = [
            f'Could not find a description for argument {key}. Please provide a description for {key}'
        ]


supported_argument_fields = [
    'default_value',
    'description',
    'key',
    'key_value_separator',
    'options',
    'required',
    'sub_arguments',
    'type',
]


def validate_unsupported_argument_fields(key, argument_data, error_dict):
    for field in argument_data.keys():
        if field not in supported_argument_fields:
            error_dict['unsupported_field'] = [
                f'The argument field {field} on {key} is not valid'
            ]



import logging
import subprocess

from docker.errors import ImageNotFound, APIError  # type: ignore

from biolib import utils
from biolib.biolib_docker_client import BiolibDockerClient
from biolib.compute_node.cloud_utils import CloudUtils
from biolib.compute_node.job_worker.cache_state import DockerImageCacheState, DockerCacheStateError
from biolib.compute_node.job_worker.cache_types import DockerImageInfo, DockerImageCacheStateDict, DockerAuthConfig


class DockerImageCache:
    def __init__(self):
        config = CloudUtils.get_webserver_config()
        self._max_cache_size = config['max_docker_image_cache_size_bytes']  # pylint: disable=unsubscriptable-object
        self._docker_client = BiolibDockerClient().get_docker_client()
        self._docker_data_dir = self._docker_client.info()['DockerRootDir']

    @property
    def _current_cache_size_on_disk(self):
        disk_usage_command = ['/bin/bash', '-c', f'du -sb {self._docker_data_dir} | cut -f1']
        if not utils.BIOLIB_IS_RUNNING_IN_ENCLAVE:
            # Need to use sudo outside of enclave
            disk_usage_command.insert(0, 'sudo')

        du_output = subprocess.run(
            disk_usage_command,
            capture_output=True,
            check=True
        ).stdout
        return int(du_output.decode().strip())

    def _clear_space_for_image(self, estimated_image_size_bytes: int, cache_state: DockerImageCacheStateDict):
        while not self._has_space_to_pull_image(estimated_image_size_bytes, cache_state):
            self._remove_least_recently_used_image(cache_state)

    def get(self, image_uri: str, estimated_image_size_bytes: int, pull_auth_config: DockerAuthConfig):
        try:
            with DockerImageCacheState() as cache_state:
                self._docker_client.images.get(image_uri)
                cache_state[image_uri]['last_used_at'] = DockerImageCacheState.get_timestamp_now()

        except ImageNotFound:
            if estimated_image_size_bytes > self._max_cache_size:
                CloudUtils.log(
                    f'Image {image_uri} with size: {estimated_image_size_bytes} is bigger than the max cache size',
                    logging.ERROR,
                )
                raise DockerCacheStateError(  # pylint: disable=raise-missing-from
                    'Image is bigger than the max cache size'
                )

            with DockerImageCacheState() as cache_state:
                if not self._has_space_to_pull_image(estimated_image_size_bytes, cache_state):
                    self._clear_space_for_image(estimated_image_size_bytes, cache_state)

                cache_state[image_uri] = DockerImageInfo(
                    last_used_at=DockerImageCacheState.get_timestamp_now(),
                    uri=image_uri,
                    state='pulling',
                    estimated_image_size_bytes=estimated_image_size_bytes
                )

            CloudUtils.log(f'Image {image_uri} not found in cache. Pulling...', logging.DEBUG)
            self._docker_client.images.pull(image_uri, auth_config=pull_auth_config)
            with DockerImageCacheState() as cache_state:
                cache_state[image_uri]['state'] = 'ready'

    def _remove_least_recently_used_image(self, cache_state: DockerImageCacheStateDict) -> None:
        cached_images = [image for image in cache_state.values() if image['state'] == 'ready']
        images_sorted_by_least_recently_used = sorted(cached_images, key=lambda image: image['last_used_at'])

        for image in images_sorted_by_least_recently_used:
            CloudUtils.log(f"Removing image: {image['uri']}", logging.DEBUG)
            try:
                self._docker_client.api.remove_image(image=image['uri'])
            except APIError as error:
                CloudUtils.log(
                    f'Could not remove image due to {error}... Skipping removal of this image.',
                    logging.ERROR
                )
                continue  # Image is in use or cannot be removed at this time

            cache_state.pop(image['uri'])
            break

    def _has_space_to_pull_image(self, estimated_image_size_bytes: int, cache_state: DockerImageCacheStateDict) -> bool:
        size_of_images_being_pulled = sum([image['estimated_image_size_bytes'] for image in cache_state.values()
                                           if image['state'] == 'pulling'])
        current_cache_size = self._current_cache_size_on_disk + size_of_images_being_pulled

        cache_space_remaining = self._max_cache_size - current_cache_size
        CloudUtils.log(f'Cache space remaining: {cache_space_remaining}', logging.DEBUG)

        return bool(cache_space_remaining > estimated_image_size_bytes)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Emulation Manager module."""
import asyncio

from biolib.pyppeteer.pyppeteer.connection import CDPSession
from biolib.pyppeteer.pyppeteer.models import Protocol


class EmulationManager:
    """EmulationManager class."""

    def __init__(self, client: CDPSession) -> None:
        """Make new emulation manager."""
        self._client = client
        self._emulatingMobile = False
        self._hasTouch = False
        self._emulatingMobile = False

    async def emulateViewport(self, viewport: Protocol.Page.Viewport) -> bool:
        """
        Evaluate viewport.
        :param viewport: dictionary which supports keys: isMobile, width, height,
            deviceScaleFactor, isLandscape, hasTouch
        """
        mobile = viewport.get('isMobile', False)
        width = viewport.get('width')
        height = viewport.get('height')
        deviceScaleFactor = viewport.get('deviceScaleFactor', 1)
        if viewport.get('isLandscape'):
            screenOrientation = {'angle': 90, 'type': 'landscapePrimary'}
        else:
            screenOrientation = {'angle': 0, 'type': 'portraitPrimary'}
        hasTouch = viewport.get('hasTouch', False)

        await asyncio.gather(
            self._client.send(
                'Emulation.setDeviceMetricsOverride',
                {
                    'mobile': mobile,
                    'width': width,
                    'height': height,
                    'deviceScaleFactor': deviceScaleFactor,
                    'screenOrientation': screenOrientation,
                },
            ),
            self._client.send('Emulation.setTouchEmulationEnabled', {'enabled': hasTouch}),
        )
        reloadNeeded = self._emulatingMobile != mobile or self._hasTouch != hasTouch
        self._emulatingMobile = mobile
        self._hasTouch = hasTouch
        return reloadNeeded



import cbor2  # type: ignore
import cose  # type: ignore

from cose import EC2, CoseAlgorithms, CoseEllipticCurves
from Crypto.Util.number import long_to_bytes
from OpenSSL import crypto

from biolib import utils
from biolib.biolib_errors import BioLibError


class NitroEnclaveUtils:
    def attest_enclave_and_get_rsa_public_key(self, expected_pcrs_and_aws_cert, attestation_document_bytes):
        cbor_data = cbor2.loads(attestation_document_bytes)
        cbor_document = cbor2.loads(cbor_data[2])
        expected_pcrs_list = expected_pcrs_and_aws_cert['pcrObjectArray']
        aws_nitro_root_cert_pem = "-----BEGIN CERTIFICATE-----\n" + \
                                  expected_pcrs_and_aws_cert['awsNitroRootCertificateBase64'] + \
                                  "\n-----END CERTIFICATE-----"
        actual_pcrs = {str(k): v.hex() for k, v in cbor_document['pcrs'].items()}

        if not utils.BIOLIB_CLOUD_SKIP_PCR_VERIFICATION:
            self._assert_pcr_validity(actual_pcrs, expected_pcrs_list)

        cert = crypto.load_certificate(crypto.FILETYPE_ASN1, cbor_document['certificate'])
        self._assert_certificate_chain_validity(cbor_document['cabundle'], cert, aws_nitro_root_cert_pem)
        self._assert_cose_signature_validity(cert, cbor_data)
        return cbor_document['public_key']

    def _assert_pcr_validity(self, actual_pcrs, expected_pcrs_list):
        for expected_pcrs in expected_pcrs_list:
            if self._does_pcr_object_match_actual(actual_pcrs, expected_pcrs):
                return True

        # Raise error if no match was fund
        raise BioLibError('Failed to verify PCRs')

    def _does_pcr_object_match_actual(self, actual_pcrs, expected_pcrs):
        for pcr_key in expected_pcrs.keys():
            if actual_pcrs[pcr_key] != expected_pcrs[pcr_key]:
                return False
            # If all matched return True
            return True

    def _assert_certificate_chain_validity(self, cabundle, cert, aws_nitro_root_cert_pem):
        store = crypto.X509Store()
        _cert = crypto.load_certificate(crypto.FILETYPE_PEM, aws_nitro_root_cert_pem)
        store.add_cert(_cert)

        for _cert_binary in cabundle[1:]:
            _cert = crypto.load_certificate(crypto.FILETYPE_ASN1, _cert_binary)
            store.add_cert(_cert)

        store_ctx = crypto.X509StoreContext(store, cert)
        try:
            store_ctx.verify_certificate()
        except Exception as exception:
            raise BioLibError('Failed to verify certificates') from exception

    def _assert_cose_signature_validity(self, cert, cbor_data):
        phdr = cbor2.loads(cbor_data[0])
        uhdr = cbor_data[1]
        signature = cbor_data[3]

        cert_public_numbers = cert.get_pubkey().to_cryptography_key().public_numbers()
        x_long = cert_public_numbers.x
        y_long = cert_public_numbers.y

        x_bytes = long_to_bytes(x_long)
        y_bytes = long_to_bytes(y_long)

        # Create the EC2 key from public key parameters
        key = EC2(alg=CoseAlgorithms.ES384, x=x_bytes, y=y_bytes, crv=CoseEllipticCurves.P_384)

        # Construct the Sign1 message
        msg = cose.Sign1Message(phdr=phdr, uhdr=uhdr, payload=cbor_data[2])
        msg.signature = signature

        # Verify the signature using the EC2 key
        if not msg.verify_signature(key):
            raise Exception('Failed to verify signature in attestation document')



import requests

from biolib.biolib_api_client.auth import BearerAuth
from biolib.biolib_api_client import BiolibApiClient
from biolib.biolib_api_client.lfs_types import LfsVersionPresignedUploadUrlResponse, LfsUploadPartMetadata, \
    LargeFileSystemVersion, LargeFileSystem, ZipProxyFileListResponse
from biolib.biolib_errors import BioLibError
from biolib.typing_utils import List


class BiolibLargeFileSystemApi:

    @staticmethod
    def create(account_uuid: str, name: str) -> LargeFileSystem:
        response = requests.post(
            f'{BiolibApiClient.get().base_url}/api/lfs/',
            auth=BearerAuth(BiolibApiClient.get().access_token),
            timeout=5,
            json={'account_uuid': account_uuid, 'name': name}
        )

        if not response.ok:
            raise BioLibError(response.content.decode())

        lfs: LargeFileSystem = response.json()
        return lfs

    @staticmethod
    def fetch_version(lfs_version_uuid: str) -> LargeFileSystemVersion:
        response = requests.get(
            f'{BiolibApiClient.get().base_url}/api/lfs/versions/{lfs_version_uuid}/',
            auth=BearerAuth(BiolibApiClient.get().access_token),
            timeout=5,
        )

        if not response.ok:
            raise BioLibError(response.content.decode())

        lfs_version: LargeFileSystemVersion = response.json()
        return lfs_version

    @staticmethod
    def fetch_file_list(presigned_download_url: str) -> ZipProxyFileListResponse:
        base_url = BiolibApiClient.get().base_url
        if base_url == 'https://biolib.com':
            source_file_proxy_url = 'https://blbcdn.com/source-file-proxy'
        elif base_url == 'https://staging.biolib.com':
            source_file_proxy_url = 'https://staging.blbcdn.com/source-file-proxy'
        else:
            source_file_proxy_url = f'{base_url}/source-file-proxy'

        response = requests.post(
            f'{source_file_proxy_url}/files/',
            timeout=10,
            json={'zip_url': presigned_download_url},
        )

        if not response.ok:
            raise BioLibError(response.content.decode())

        response_dict: ZipProxyFileListResponse = response.json()
        return response_dict

    @staticmethod
    def create_version(resource_uuid: str) -> LargeFileSystemVersion:
        response = requests.post(
            f'{BiolibApiClient.get().base_url}/api/lfs/versions/',
            auth=BearerAuth(BiolibApiClient.get().access_token),
            json={'resource_uuid': resource_uuid},
            timeout=5,
        )

        if not response.ok:
            raise BioLibError(response.content.decode())

        lfs_version: LargeFileSystemVersion = response.json()
        return lfs_version

    @staticmethod
    def get_upload_url(resource_version_uuid: str, part_number: int) -> LfsVersionPresignedUploadUrlResponse:
        response = requests.get(
            f'{BiolibApiClient.get().base_url}/api/lfs/versions/{resource_version_uuid}/presigned_upload_url/',
            auth=BearerAuth(BiolibApiClient.get().access_token),
            params={'part_number': part_number},
            timeout=5,
        )

        if not response.ok:
            raise BioLibError(response.content.decode())

        presigned_upload_url_response: LfsVersionPresignedUploadUrlResponse = response.json()
        return presigned_upload_url_response

    @staticmethod
    def complete_upload(
            resource_version_uuid: str,
            parts: List[LfsUploadPartMetadata],
            size_bytes: int,
    ) -> None:
        response = requests.post(
            f'{BiolibApiClient.get().base_url}/api/lfs/versions/{resource_version_uuid}/complete_upload/',
            auth=BearerAuth(BiolibApiClient.get().access_token),
            json={'parts': parts, 'size_bytes': size_bytes},
            timeout=5,
        )

        if not response.ok:
            raise BioLibError(response.content.decode())



import time

from biolib import utils
from biolib.compute_node.cloud_utils import CloudUtils
from biolib.typing_utils import Dict, List
from biolib.compute_node.webserver.worker_thread import WorkerThread


JOB_ID_TO_COMPUTE_STATE_DICT: Dict = {}
UNASSIGNED_COMPUTE_PROCESSES: List = []


def finalize_and_clean_up_compute_job(job_id: str):
    JOB_ID_TO_COMPUTE_STATE_DICT.pop(job_id)
    if utils.IS_RUNNING_IN_CLOUD:
        CloudUtils.finish_job(job_id)


def get_compute_state(unassigned_compute_processes):
    if len(unassigned_compute_processes) == 0:
        start_compute_process(unassigned_compute_processes)

    return unassigned_compute_processes.pop()


def start_compute_process(unassigned_compute_processes):
    compute_state = {
        'job_id': None,
        'status': {
            'status_updates': [
                {
                    'progress': 10,
                    'log_message': 'Initializing'
                }
            ],
        },
        'result': None,
        'attestation_document': b'',
        'received_messages_queue': None,
        'messages_to_send_queue': None,
        'worker_process': None
    }

    WorkerThread(compute_state).start()

    while True:
        if compute_state['attestation_document']:
            break
        time.sleep(0.25)

    unassigned_compute_processes.append(compute_state)


def validate_saved_job(saved_job):
    if 'app_version' not in saved_job['job']:
        return False

    if 'access_token' not in saved_job:
        return False

    if 'module_name' not in saved_job:
        return False

    return True
